header 
{
package jburg.parser;

}

{
import java.io.*;

import antlr.CommonAST;
import antlr.collections.AST;
}

class JBurgParser extends Parser; 
options {
   buildAST = true;         // uses CommonAST by default
   k=7;                     
   exportVocab=JBurg;
}


tokens 
{ 
    BURM_PROPERTY = "BURMProperty";
    DEFAULT_ERROR_HANDLER = "DefaultErrorHandler";
    HEADER = "header";
    INODE_ADAPTER = "INodeAdapter";
    INODE_TYPE = "INodeType";
    IMPLEMENTS = "implements";
    LANGUAGE = "Language";
    OPCODE_TYPE = "OpcodeType";
    PACKAGE = "package";
    PATTERN = "Pattern";
    PROLOGUE = "Prologue";
    RETURN_TYPE = "ReturnType";
    VOID = "void";

    //  Synthetic token types used to synthesize nodes.
	COST_FUNCTION; 
	FUNCTION_CALL; 
	HEADER_DECLARATION;
	IMPLEMENTS_INTERFACE_SPECIFICATION;
	INCLASS_DECLARATION;
	INODE_ADAPTER_DECLARATION;
	INODE_TYPE_DECLARATION;
	LANGUAGE_DECLARATION;
	NON_TERMINAL_PARAMETER;
	OPERATOR_SPECIFICATION;
	OPERAND_LIST;
	OPERAND_AT_LEAST_ONE_ARITY;
	OPERAND_ARBITRARY_ARITY;
	PACKAGE_SPECIFICATION; 
    PATTERN_DECLARATION;
	PATTERN_RULE; 
    PATTERN_SPECIFICATION;
	PROPERTY_SPECIFICATION; 
    REDUCTION_ACTION;
    REDUCTION_DECLARATION;
	RETURN_DECLARATION;
	SIMPLE_COST_SPEC; 
	SIMPLE_TRANSFORMATION_RULE; 
	TRANSFORMATION_RULE; 
	TERMINAL_PATTERN; 
	TYPED_RETURN_DECLARATION;
}

/*
 *  Support routines to be included in the parser.
 */
{
    /*
     *  Error-handling code.
     */
    boolean parseOk = true;

	public void reportError(antlr.RecognitionException e)
	{
        super.reportError(e);
        parseOk = false;
	}

    public boolean parseSuccessful()
    {
        return this.parseOk;
    }
}

/*
 *  **********************************************************
 *  **  Parser productions (mostly in alphabetical order).  **
 *  **********************************************************
 */

//  "start" production is specification.

cost_function!
   :    funcId:IDENTIFIER LPAREN RPAREN functionBody:BLOCK 
   {
    #cost_function = #([COST_FUNCTION], funcId, functionBody);
   }
   ;

cost_specification
   :    COLON! ( simple_cost_spec | function_call )
   ;

declaration
   :   error_handler_declaration
   |   implements_declaration
   |   language_declaration
   |   header_declaration
   |   inode_type_declaration
   |   inode_adapter_declaration
   |   package_declaration
   |   property_specification
   |   inclass_declaration
   |   return_declaration
   |   opcode_declaration
   ;

error_handler_declaration!
    :   DEFAULT_ERROR_HANDLER action:BLOCK
    {
        #error_handler_declaration  = #([DEFAULT_ERROR_HANDLER], action);
    }
    ;

function_call!
   :    funcId:IDENTIFIER LPAREN RPAREN
   {
    #function_call = #([FUNCTION_CALL], funcId );
   }
   ;

header_section
   :    ( declaration )+
   ;

header_declaration!
   :    HEADER header_block:BLOCK
   {
    #header_declaration = #([HEADER_DECLARATION], header_block);
   }
   ;
   
inclass_declaration!
	: inclass_block:BLOCK
	{
	 #inclass_declaration = #([INCLASS_DECLARATION], inclass_block);
	}
	;

inode_adapter_declaration!
   :    INODE_ADAPTER inode_adapter_class:multipart_identifier SEMI
   {
    #inode_adapter_declaration = #([INODE_ADAPTER_DECLARATION], inode_adapter_class);
   } 
   ;
   
inode_type_declaration!
   :    INODE_TYPE inode_type:multipart_identifier SEMI
   {
    #inode_type_declaration = #([INODE_TYPE_DECLARATION], inode_type);
   } 
   ;
   
language_declaration!
   :    LANGUAGE lang_name:IDENTIFIER SEMI
   {
   	#language_declaration = #([LANGUAGE_DECLARATION], lang_name);
   }
   ;

implements_declaration!
   :    IMPLEMENTS implements_interface_name:multipart_identifier SEMI
   {
    #implements_declaration = #([IMPLEMENTS_INTERFACE_SPECIFICATION], implements_interface_name);
   }
   ;

multipart_identifier
   :    firstPart:IDENTIFIER ( ( PERIOD | (COLON COLON) ) IDENTIFIER )* (STAR)?
   {
	   //  Coalesce all the child nodes' text.
       StringBuffer allText = new StringBuffer();

        AST current = #firstPart;

        while (current != null) 
		{
            allText.append(current.getText());
            current = current.getNextSibling();
		}

	   //  Synthesize an IDENTIFIER node, and
       //  give it the re-assembled multipart ID.
       #multipart_identifier = #([IDENTIFIER]);
	   #multipart_identifier.setText( allText.toString() );
   }
   ;

multipart_identifier_java_only
   :    firstPart:IDENTIFIER ( PERIOD IDENTIFIER )?
   {
	   //  Coalesce all the child nodes' text.
       StringBuffer allText = new StringBuffer();

        AST current = #firstPart;

        while (current != null) 
		{
            allText.append(current.getText());
            current = current.getNextSibling();
		}

	   //  Synthesize a node with an explict type.
       #multipart_identifier_java_only = #([IDENTIFIER]);
	   #multipart_identifier_java_only.setText( allText.toString() );
   }
   ;

opcode_declaration!
   :    OPCODE_TYPE opcodeType:multipart_identifier_java_only SEMI
   {
       #opcode_declaration = #([OPCODE_TYPE], #opcodeType);
   }
   ;
operand_list
   :    operand (COMMA! operand)* (COMMA! n_ary_operand)?
   {
       #operand_list = #([OPERAND_LIST], #operand_list);
   }
   |    manyop:n_ary_operand
   {
       #operand_list = #([OPERAND_LIST], manyop );
   }
   ;

operand
   :
    operator_specification      //  Compositional/fuctional style
   |    parameter_decl              //  Simple list of subgoals
   ;

operator_specification!
   :
       ( multipart_identifier_java_only LPAREN VOID RPAREN ) =>
       terminalId:multipart_identifier_java_only LPAREN VOID RPAREN
   {
       #operator_specification = #([TERMINAL_PATTERN], terminalId);
   }
   |   
       ( multipart_identifier_java_only LPAREN VOID RPAREN IDENTIFIER) =>
       namedTerminalId:multipart_identifier_java_only LPAREN VOID RPAREN terminalName:IDENTIFIER
   {
       #operator_specification = #([TERMINAL_PATTERN], namedTerminalId, terminalName);
   }
   |    
       operatorId:multipart_identifier_java_only LPAREN operands:operand_list RPAREN
   {
       #operator_specification = #([OPERATOR_SPECIFICATION], operatorId, operands);
   }
   ;

n_ary_operand!
   :    at_least_one_operand:parameter_decl PLUS!  //  Number of operands may vary, but at least one
   {
       #n_ary_operand = #([OPERAND_AT_LEAST_ONE_ARITY], at_least_one_operand );
   }
   |    arbitrary_operands:parameter_decl STAR!  //  Number of operands may vary
   {
       #n_ary_operand = #([OPERAND_ARBITRARY_ARITY], arbitrary_operands );
   }
   ;

package_declaration!
   :    PACKAGE package_name:multipart_identifier SEMI
   {
    #package_declaration = #([PACKAGE_SPECIFICATION], package_name);
   }
   ;

parameter_decl!
   : id1:IDENTIFIER id2:IDENTIFIER
   	{
	    #parameter_decl = #([NON_TERMINAL_PARAMETER], id1, id2);
	}
   ;

pattern_declaration!
    :   PATTERN name:IDENTIFIER pattern:pattern_specification
        {
            #pattern_declaration = #([PATTERN_DECLARATION], name, pattern);
        }

        SEMI
    ;

pattern_specification!
	: op:operator_specification
	{
		#pattern_specification = #([PATTERN_SPECIFICATION], op);
	}
	;

property_specification!
   :  BURM_PROPERTY property_type:type_name property_name:IDENTIFIER SEMI
   {
    #property_specification = #([PROPERTY_SPECIFICATION], property_type, property_name);
   }
   ;

return_declaration!
   :    RETURN_TYPE return_type:type_name SEMI
   {
    #return_declaration = #([RETURN_DECLARATION], return_type);
   } 
   |    RETURN_TYPE state_name:IDENTIFIER EQUALS state_return_type:type_name SEMI
   {
    #return_declaration = #([TYPED_RETURN_DECLARATION], state_name, state_return_type);
   }
   ;

/**
 *  Reduce a rule.
 *  Rule types are:<ul>
 *  <li>  Pattern rules, which combine subgoals with other nodes further up the tree.
 *  <li>  Simple transformation rules.  Like iBurg's chain rules, simple transformation
 *        rules allow one goal to satisfy additional goals.
 *  <li>  Complex transformation rules, which allow the reducer to actively transform
 *        a subgoal to satisfy additional subgoals.
 *  <li>  A leaf node, with no children, can be used as a subgoal in either type 
 *        of transformation rule.
 *  </ul>
 */
rule!
   :    nonTerminalRuleId:IDENTIFIER EQUALS pattern:pattern_specification cost:cost_specification action:reduction_action    
   {
        #rule = #([PATTERN_RULE], nonTerminalRuleId, pattern, cost, action);
   }
   |    simpleTransformationTarget:IDENTIFIER 
    EQUALS! simpleTransformationSource:IDENTIFIER 
    SEMI!
   {
    #rule = #([SIMPLE_TRANSFORMATION_RULE], simpleTransformationTarget, simpleTransformationSource );
   }
   |    transformationTarget:IDENTIFIER 
    EQUALS!  transformationSource:IDENTIFIER 
    transformationCost:cost_specification 
    transformationAction:reduction_action
   {
	#rule = #([TRANSFORMATION_RULE], transformationTarget, transformationSource, transformationCost, transformationAction );
   }
   ;

reduction_action
    : (PROLOGUE prologue:BLOCK)? action:BLOCK
    {
        #reduction_action = #([REDUCTION_ACTION], action);
        if ( prologue != null )
            #reduction_action.addChild(#([PROLOGUE],prologue));
    }
    ;

reduction_declaration!
    :   nonTerminalRuleId:IDENTIFIER EQUALS  PATTERN name:IDENTIFIER cost:cost_specification r:reduction_action
    {
        #reduction_declaration = #([REDUCTION_DECLARATION], nonTerminalRuleId, name, cost, r);
    }
    ;

simple_cost_spec!
   :    iCost:INT
   {
       #simple_cost_spec = #([SIMPLE_COST_SPEC], iCost);
   }
   ;

specification
   :    header_section (rule | cost_function | pattern_declaration | reduction_declaration )+ EOF!
   ;

type_name!
    : stem:multipart_identifier (LANGLE ptype:multipart_identifier RANGLE)?
    {
        if ( null == #ptype )
        {
            #type_name = #stem;
        }
        else
        {
            StringBuffer buffer = new StringBuffer(#stem.getText());
            buffer.append("<");
            buffer.append(#ptype.getText());
            buffer.append(">");
            #type_name = #([IDENTIFIER]);
            #type_name.setText(buffer.toString());
        }
    }
    ;


/*
 *  ***************************************
 *  **  Lexical analyzer specification.  **
 *  ***************************************
 */
class JBurgANTLRLexer extends Lexer;
options
{
   //  Accept ASCII characters as input; UNICODE would be great but too many alternatives breaks java switch stmt
   charVocabulary = '\3'..'\377';
   // charVocabulary = '\3'..'\377' | '\u1000'..'\u1fff';

   k=8;
}

{

    /**  
     *  Path to the main source file so that includes can be
     *  relative to it.
     */
    public java.io.File mainSourceFile = null;

    /*
     *  Multiple token stream support for include files.
     */
    public static antlr.TokenStreamSelector selector; // must be assigned externally

    public void uponEOF() throws TokenStreamException, CharStreamException
    {
        try
        {
            selector.pop(); // return to old lexer/stream
            selector.retry();
        }
        catch (java.util.NoSuchElementException e)
        {
            // No including stream, return EOF
        }
    }

}

WS :    (' '
   |    '\t'
   |    '\n'    {newline();}
   |    '\r')
    { _ttype = Token.SKIP; }
   ;

COLON:  ':';
COMMA:  ',';
EQUALS: '=';
LANGLE: '<';
LPAREN: '(';
PERIOD: '.';
PLUS:   '+';
RANGLE: '>';
RPAREN: ')';
SEMI:   ';';
STAR:   '*';

// Single-line comments
COMMENT
  : "//" (~('\n'|'\r'))* { $setType(Token.SKIP);}
  ;   


// multiple-line comments
ML_COMMENT
  : "/*"
    ( { LA(2)!='/' }? '*'
    | ( ("\r\n") => "\r\n" |'\n') { newline(); }
    | ~('*'|'\n'|'\r')
    )*
    "*/"
    { $setType(Token.SKIP); }
  ;   
  
protected
DIGIT
   :    '0'..'9'
   ;

INT :   ('-')? (DIGIT)+
   ;

protected DIRECTIVE_STRING
    : '"'! ( '\\' . | ~('\\'|'"') )* '"'!
    //  Hokey: re-sync syntax coloring "
    ;

INCLUDE_DIRECTIVE
    : "JBurg.include" (WS)* fileName:DIRECTIVE_STRING
    {
        String includeFileName = fileName.getText();
        java.io.File includeFile = null;

        try
        {
            //  Try to open the include file relative to the main source file.
            includeFile = new java.io.File(mainSourceFile.getParentFile(), includeFileName);
            if ( !includeFile.exists() )
                includeFile = new java.io.File(includeFileName);

            //  Go ahead and try to use it as-is; the FileNotFoundException has
            //  to be caught from API calls, so let it do error handling.
            JBurgANTLRLexer sublexer = new JBurgANTLRLexer(new java.io.DataInputStream(new java.io.FileInputStream(includeFile)));
            sublexer.mainSourceFile = includeFile;
            sublexer.setFilename(includeFile.getName());
            selector.push(sublexer);
            selector.retry();
        }
        catch ( java.io.FileNotFoundException no_file )
        {
            System.err.println("Unable to include " + includeFile + " : " + no_file.getMessage());
            //  Without a retry() call this will cascade into a parse error.
        }
    }
    ;
ECHO_DIRECTIVE
    : "#jburg.echo" (WS)* message:DIRECTIVE_STRING
    {
        System.out.println("jburg.echo: " + message.getText());
        $setType(Token.SKIP);
    }
    ;

IDENTIFIER
options {
   testLiterals = true;
   paraphrase = "an identifier";
}
   :    ('a'..'z' | 'A'..'Z') ('a'..'z'|'A'..'Z'|'_'|'0'..'9')*
   ;


BLOCK
options
{
   paraphrase = "a block of code";
}
   :   '{'  
           ( BLOCK 
           | ( ("\r\n") => "\r\n" |'\n') { newline(); }
           | ~( '{' | '}'|'\n'|'\r')
           )* 
       '}'
   ;   

